m4_changequote([,])m4_ifelse([

2001-06-08:

Macros for creating scalar cache. Scalar, in this context, means that
we cache only one set of function parameters, even though the result
type might be an array of constant shape.

Note that, due to the access pattern induced by the node lists, a cache
for a density-dependent function will never be used.

Furthermore, note that the use of variables with the SAVE attribute
necessary for maintaining the cache may adversely affect optimization
(esp. on a lousy compiler, like NAG's) and thus actually lead to an
increase in computing time. Even (much) more importantly, if you
decide to test equality of floating point numbers a and b via
"abs(a-b) < spacing(a)" or the likes of it instead of a simple "a==b",
the penalty of branch prediction and evaluation of the criterion may
again slow things down considerably. Therefore, NEVER use caching
without testing the its effect.

Use these macros like this (but note that, in reality, you should omit
the expensive call to realEqual (v.s.), even though this may produce
some compiler warnings):

    #public #real function f(a,b,c,i,j)
    #real, intent(in) :: a,b,c
    integer, intent(in) :: i,j
    makeScalarCache__resultType__name_type_eq_(`#real',
       a,`#real',realEqual($1,$2),
       b,`#real',realEqual($1,$2),
       c,`#real',realEqual($1,$2),
       i,integer,,
       j,integer,)
    <-- Code for calculating f goes here. -->
    #end

Depending upon m4 macro m4_ifCache [ctp] a cache is created. If so,
the above call to makeScalarCache__... introduces additional variables
cacheValid, cacheResult, cached_a, cached_b, cached_c, cached_i and
cached_j. Note that you only have to add one macro call in the front
of the productive code; this is possible by using the automatic
...:auto-pre-end:...-hook; any previous setting for this hook will be
inserted after the modifications induced by creating the cache.

])m4_changequote`'m4_dnl
m4_dnl 
m4_dnl Process ctp __useCache__:
m4_dnl 
m4_ifdef(`__useCache__',,`m4_define(`__useCache__',`yes')')m4_dnl
m4_checkrange(`__useCache__', `yes', `no')m4_dnl
m4_ifelse(__useCache__, `yes',
  `m4_define(`m4_ifCache', `$1')',
  `m4_define(`m4_ifCache', `$2')')m4_dnl
m4_dnl 
m4_dnl Honor m4 macro `m4_ifCache' [ctp] for defining the macros:
m4_dnl 
m4_ifCache(`
m4_dnl 
m4_dnl 
m4_dnl auxiliary macros:
m4_dnl 
m4_dnl print out series of .and.-equality-test-statements;
m4_dnl declare caching variables
m4_dnl 
m4_define(`_makeScalarCache_aux__name_type_eq_rest',
  `m4_ifelse($#,0,,
    `m4_ifelse(m4_ifelse($#,1,e,$#,2,e),e,`
       --error: $# arguments to macro
       --error: ``$0''
       --error: while making scalar cache on fun`'ction m4_proc
       --error: (3 arguments expected).
     ',
      ` .and. &m4_pushdef(`_aux_', m4_ifelse(`$3',,`$'`1 == $'`2', `$3'))
       _aux_(cached_$1, $1)`'m4_popdef(
         `_aux_').._PRIVATE_VARS_>>$2, save :: cached_$1
       <<_PRIVATE_VARS_..`'m4_ifelse(
         $#,3,,`$0(m4_shift(m4_shift(m4_shift($@))))')')')')m4_dnl
m4_dnl 
m4_dnl cache old parameter settings
m4_dnl 
m4_define(`_makeScalarCache_AUX__name_type_eq_rest',
  `m4_ifelse($#,0,,
    `m4_ifelse(m4_ifelse($#,1,e,$#,2,e),e,`
       --error: $# arguments to macro
       --error: ``$0''
       --error: while making scalar cache on fun`'ction m4_proc
       --error: (3 arguments expected).
     ',
      `cached_$1 = $1<:="\n":>m4_dnl
       m4_ifelse($#,3,,`$0(m4_shift(m4_shift(m4_shift($@))))')m4_dnl
  ')')')m4_dnl
m4_dnl ************** main macro for scalar cache
m4_dnl 
m4_dnl arguments: result type of function
m4_dnl            for every input parameter:
m4_dnl              name, type, equality test
m4_dnl 
m4_dnl The equality test is either in the form of a string expanding to
m4_dnl a macro definition, or empty (then taken to be a simple application
m4_dnl of == in Fortran.)
m4_dnl 
m4_define(`makeScalarCache__resultType__name_type_eq_', `
  .._PRIVATE_VARS_>>m4_dnl
    logical, save :: cacheValid = .false.
    $1, save :: cacheResult
  <<_PRIVATE_VARS_..
  if (cacheValid m4_dnl .and. arguments == cached arguments
      _makeScalarCache_aux__name_type_eq_rest(m4_shift($@)) ) then
    <:=$result:> = cacheResult
m4_ifdef(hook m4_sourcefilebase:m4_proc:auto-cache-used,
`! begin hook m4_sourcefilebase:m4_proc:auto-cache-used
m4_indir(hook m4_sourcefilebase:m4_proc:auto-cache-used)
! end hook m4_sourcefilebase:m4_proc:auto-cache-used:
')..HOOKS>>hook m4_sourcefilebase:m4_proc:auto-cache-used\n<<HOOKS..m4_dnl
  else m4_dnl
m4_ifdef(hook m4_sourcefilebase:m4_proc:auto-cache-not-used,
`! begin hook m4_sourcefilebase:m4_proc:auto-cache-not-used
m4_indir(hook m4_sourcefilebase:m4_proc:auto-cache-not-used)
! end hook m4_sourcefilebasem4_proc:auto-cache-not-used:
')..HOOKS>>hook m4_sourcefilebase:m4_proc:auto-cache-not-used\n<<HOOKS..m4_dnl
    m4_define(`hook 'm4_sourcefilebase:m4_unitname`:auto-pre-end:'m4_proc,
      cacheValid = .true.<:="\n":>m4_dnl
      _makeScalarCache_AUX__name_type_eq_rest(m4_shift($@))m4_dnl
      cacheResult = <:=$result:>
    end if
    m4_defn(`hook 'm4_sourcefilebase:m4_unitname`:auto-pre-end:'m4_proc)
  )m4_dnl
')m4_dnl
',`m4_define(`makeScalarCache__resultType__name_type_eq_')')m4_dnl
