m4_dnl                 PROPERTIES AND CAPABILITIES:
m4_dnl                 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
m4_dnl 
m4_dnl In order to overcome the problems with hook definitions valid only for
m4_dnl certain combinations of main parts, we install an additional protocol:
m4_dnl One main part may export a list of properties, and if those correspond
m4_dnl to the capabilities of another main part the latter may generate
m4_dnl customized code.
m4_dnl 
m4_dnl ________________________________________________________________
m4_dnl 
m4_dnl PROPERTIES:
m4_dnl ~~~~~~~~~~~
m4_dnl 
m4_dnl (*) Add a (possibly new) property. This simply maintains a list of
m4_dnl known properties (`properties m4_sourcefilebase'), successively
m4_dnl prepends the parameter-value-pairs to the definition of a
m4_dnl property-specific macro, `property m4_sourcefilebase ...').  As a
m4_dnl consequence, the order of the sequence-value pairs as well as of the
m4_dnl individual instances of the property are reversed. Different parameter
m4_dnl sets for the same property type are separated by a single empty list
m4_dnl entry. The list of properties has a trailing comma.
m4_dnl 
m4_define(`newProperty__property__arg_val___',
 `m4_ifdef(`property 'm4_sourcefilebase` $1',
    `m4_define(`property 'm4_sourcefilebase` $1',
               `,'m4_defn(`property 'm4_sourcefilebase` $1'))',
    `m4_define(`properties 'm4_sourcefilebase,
       `$1,'m4_defn(`properties 'm4_sourcefilebase))')m4_ifelse(
 )m4_pushdef(`register',
    m4_changequote([,])[m4_ifelse(
     )m4_ifelse($][#, 1, ,
                $][#, 2, ,
                `m4_define(`$][1', `$][2,$][3,'m4_defn($][1))m4_ifelse(
                 )register(`$][1', m4_shift(m4_shift(m4_shift($][@))))')]m4_dnl
    m4_changequote)m4_dnl
  register(`property 'm4_sourcefilebase` $1', m4_shift($@))m4_dnl
  m4_popdef(`register')')
m4_dnl 
m4_dnl 
m4_dnl (*) Dump definitions of properties
m4_dnl 
m4_dnl Output definitions of all properties of the current main part. This
m4_dnl relies on the list `properties 'm4_sourcefilebase,
m4_dnl cf. newProperty_... .
m4_dnl 
m4_define(`dumpProperties',
 `m4_ifdef(`properties 'm4_sourcefilebase,
   `m4_pushdef(`__dump__', m4_changequote([,
     ])[m4_ifelse(`$][1',,,
        `m4_changequote(|m4:,:m4|)
         |m4:m4_define(`property :m4|m4_sourcefilebase|m4: $][1',
          `:m4|m4_defn(
              |m4:property :m4|m4_sourcefilebase|m4: $][1:m4|)|m4:'):m4|
         m4_changequote
         $][0(m4_shift($][@))')]m4_changequote)
    __dump__(m4_indir(`properties 'm4_sourcefilebase))
    m4_popdef(`__dump__')
  ')
')
m4_dnl 
m4_dnl ________________________________________________________________
m4_dnl 
m4_dnl ITERATION OVER PROPERTIES:
m4_dnl ~~~~~~~~~~~~~~~~~~~~~~~~~~
m4_dnl 
m4_dnl In simple cases we don't need the full capabilities-mechanism;
m4_dnl instead, it is sufficient to simply apply some code to the parameter
m4_dnl settings:
m4_dnl 
m4_dnl (*) Macros to iterate over parameter sets:
m4_dnl 
m4_dnl For each of the parameter sets associated with a given property ($1,
m4_dnl with prepended main part name) we call the code ($2) with the
m4_dnl corresponding key-value pairs. Note that the accumulation of the
m4_dnl arguments restores these to the original order in which they were
m4_dnl given within any parameter set, whereas the order of the parameter
m4_dnl sets is not changed and thus is the reverse of the order in which they
m4_dnl were specified.
m4_dnl 
m4_dnl For convenience, we make available the auxiliary macro getValue__: it
m4_dnl takes a key and a sequence of key-value pairs, and it returns the
m4_dnl first value found, or `' otherwise. We take care not to add any
m4_dnl whitespace. -- As getValue__ will usually be called with the current
m4_dnl set of parameters, maro getValue calls getValue__, appending these
m4_dnl arguments.
m4_dnl 
m4_dnl It is an error for a property to not be defined.
m4_dnl 
m4_define(`forEachPropertySet__MPprop_code',
 `m4_pushdef(`getValue__',
     m4_changequote([,])[m4_ifelse($][#, 2, `',
                                   `$][1', `$][2', `$][3',
                                   `$][0(`$][1', m4_shift(m4_shift(m4_shift(
                                         $][@))))')]m4_changequote)m4_ifelse(
 )m4_pushdef(`getValue', m4_changequote([,
             ])[getValue__($][@,__args__)]m4_changequote)m4_ifelse(
 )m4_define(`__code__',
    `$2`'m4_define(`__args__')')m4_pushdef(`__args__')m4_ifelse(
 )m4_pushdef(`__aux__', m4_changequote([,
    ])[m4_ifelse(
        `$][1', `', `__code__(__args__)m4_ifelse(
                        m4_eval($][#<2), 1,,`$][0(m4_shift($][@))')',
        `m4_define(`__args__',
              `$][1,$][2,'m4_defn(`__args__'))m4_ifelse(
              )$][0(m4_shift(m4_shift($][@)))')]m4_changequote)m4_ifelse(
 )__aux__(m4_indir(`property $1'))`'m4_ifelse(
 )m4_popdef(`getValue__')m4_popdef(`getValue')m4_ifelse(
 )m4_popdef(`__code__')m4_popdef(`__args__')m4_popdef(`__aux__')m4_dnl
')
m4_dnl 
m4_dnl We also provide code for indirect call of already defined macros:
m4_dnl 
m4_define(`forEachPropertySet__MPprop_macro',
 `forEachPropertySet__MPprop_code(`$1', m4_defn(`$2'))')
m4_dnl 
m4_dnl 
m4_dnl (*) Macro to apply code to most recent set of values
m4_dnl 
m4_dnl Call code on the last property set specified (which is the first item
m4_dnl in the list); this is basically the same as forEachProp..., and
m4_dnl getValue is also provided.
m4_dnl 
m4_dnl [Note that it would be easier to simply call the code with the whole
m4_dnl list, but that would prevent use of $#.]
m4_dnl 
m4_define(`forLastPropertySet__MPprop_code',
 `m4_pushdef(`getValue__',
     m4_changequote([,])[m4_ifelse($][#, 2, `',
                                   `$][1', `$][2', `$][3',
                                   `$][0(`$][1', m4_shift(m4_shift(m4_shift(
                                           $][@))))')]m4_changequote)m4_ifelse(
 )m4_pushdef(`getValue', m4_changequote([,
             ])[getValue__($][@,__args__)]m4_changequote)m4_ifelse(
 )m4_define(`__code__',
    `$2')m4_pushdef(`__args__')m4_ifelse(
 )m4_pushdef(`__aux__', m4_changequote([,
    ])[m4_ifelse(
        `$][1', `', `__code__(__args__)',
        `m4_define(`__args__',
              `$][1,$][2,'m4_defn(`__args__'))m4_ifelse(
              )$][0(m4_shift(m4_shift($][@)))')]m4_changequote)m4_ifelse(
 )__aux__(m4_indir(`property $1'))`'m4_ifelse(
 )m4_popdef(`getValue__')m4_popdef(`getValue')m4_ifelse(
 )m4_popdef(`__code__')m4_popdef(`__args__')m4_popdef(`__aux__')m4_dnl
')
m4_dnl 
m4_dnl We also provide code for indirect call of already defined macros:
m4_dnl 
m4_define(`forLastPropertySet__MPprop_macro',
 `forLastPropertySet__MPprop_code(`$1', m4_defn(`$2'))')
m4_dnl 
m4_dnl ________________________________________________________________
m4_dnl 
m4_dnl CAPABILITIES:
m4_dnl ~~~~~~~~~~~~~
m4_dnl 
m4_dnl Where the above functions are not sufficient, we have to go to
m4_dnl capabilities.  These are here implemented in a stack-like structure,
m4_dnl and we can also dispatch into a set of similarly-named capabilities.
m4_dnl 
m4_dnl (*) Stack of capability sets
m4_dnl 
m4_dnl The situation may occur when we want the same property to be treated
m4_dnl differently in different parts of the same main part. In order to
m4_dnl facilitate that, we provide a stack for capabilities.
m4_dnl 
m4_dnl - capabilities are always stored with m4_pushdef;
m4_dnl 
m4_dnl - we maintain a list of the capabilities defined at any one instance:
m4_dnl   *capabilities*;
m4_dnl 
m4_dnl - we provide m4_pushdef- / m4_popdef-like commands for the list of
m4_dnl   capabilities
m4_dnl 
m4_dnl - when dropping the list of capabilities, the corresponding
m4_dnl   capabilities are dropped, too
m4_dnl 
m4_dnl Note that this means that capabilities are implicitly inherited until
m4_dnl replaced. Also, multiple definitions of the same capability result in
m4_dnl duplicate entries in *capabilities* in order to get the correct number
m4_dnl of m4_popdef's.
m4_dnl 
m4_dnl - Push *capabilities*:
m4_dnl 
m4_define(`pushCapabilities', `m4_pushdef(`*capabilities*')')
m4_dnl 
m4_dnl - Pop *capabilities* and capabilities:
m4_dnl 
m4_define(`popCapabilities',
 `m4_pushdef(`__aux__', m4_changequote([,])[m4_ifelse(`$][1',,,
          `m4_popdef(`capability $][1')$][0(m4_shift($][@))')]m4_changequote)
  __aux__(m4_indir(`*capabilities*'))
  m4_popdef(`*capabilities*')m4_popdef(`__aux__')')
m4_dnl 
m4_dnl - Add capability:
m4_dnl 
m4_define(`newCapability__MPcap_code',
 `m4_pushdef(`capability $1', `$2')m4_ifelse(
 )m4_define(`*capabilities*', `$1,'m4_defn(`*capabilities*'))')
m4_dnl 
m4_dnl 
m4_dnl (*) Dispatch capabilities, iterating over parameter sets
m4_dnl 
m4_dnl The capabilities should have names of the form `capability mainpart
m4_dnl property:item:kind', where item is any of the values of one of the
m4_dnl fields of the property; kind can be used to distinguish between
m4_dnl different invocations of the capabilities mechanism on the same set of
m4_dnl properties. This macro will generate indirect calls to the
m4_dnl capabilities, dispatching on property (with prepended main part name)
m4_dnl and field; if the capability is not defined, an optional default value
m4_dnl may be returned.
m4_dnl 
m4_define(`dispatchCapabilities__MPprop_field_kind_default',
 `forEachPropertySet__MPprop_code(`$1',
   `m4_ifdef(`capability $1:'getValue(`$2')`:$3',
      `m4_indir(`capability $1:'getValue(`$2')`:$3', __args__)',`$4')')')
m4_dnl 
m4_dnl 
m4_dnl (*) Check whether all parameter sets can successfully be dispatched
m4_dnl 
m4_dnl Call with property (with prepended main part name) and field name as
m4_dnl arguments; expands to `+' if all calls can be dispatched, to `-' if
m4_dnl some calls cannot be dispatched, and to `?' if the property is not
m4_dnl defined at all.
m4_dnl 
m4_define(`checkCapabilities__MPprop_field_kind',
 `m4_ifelse(m4_len(forEachPropertySet__MPprop_code(
                    `$1',
                    `m4_ifdef(`capability $1:'getValue(`$2')`:$3',,`n')')),
            0, `+', `-')')
